!  omegalib.nw --
!
!  Copyright (C) 1999-2024 by
!      Wolfgang Kilian <kilian@physik.uni-siegen.de>
!      Thorsten Ohl <ohl@physik.uni-wuerzburg.de>
!      Juergen Reuter <juergen.reuter@desy.de>
!      with contributions from
!      Fabian Bach <fabian.bach@t-online.de>
!      Bijan Chokoufe Nejad <bijan.chokoufe@desy.de>
!      Christian Speckner <cnspeckn@googlemail.com>     
!
!  WHIZARD is free software; you can redistribute it and/or modify it
!  under the terms of the GNU General Public License as published by
!  the Free Software Foundation; either version 2, or (at your option)
!  any later version.
!
!  WHIZARD is distributed in the hope that it will be useful, but
!  WITHOUT ANY WARRANTY; without even the implied warranty of
!  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!  GNU General Public License for more details.
!
!  You should have received a copy of the GNU General Public License
!  along with this program; if not, write to the Free Software
!  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
module omega_birdtracks
  use kinds
  implicit none
  private
  public :: color_state
  type :: color_state
     integer, dimension(:), allocatable :: i, o
     logical :: g = .false.
  contains
     procedure :: check => color_state_check
     procedure :: to_string => color_state_to_string
     procedure :: conjugate => color_state_conjugate
     procedure :: rank_inflowing, rank_outflowing
     procedure :: contains_epsilons => color_state_contains_epsilons
  end type color_state
  integer, parameter, public :: CS_OK = 0
  integer, parameter, public :: CS_GHOST = -1
  integer, parameter :: IO_BUFFER_SIZE = 1000
  public :: epsilon3
  type :: epsilon3
     integer, dimension(3) :: j
  contains
     procedure :: check => epsilon3_check
     procedure :: to_string => epsilon3_to_string
  end type epsilon3
  integer, parameter, public :: E3_OK = 0
  integer, parameter, public :: E3_NON_NEGATIVE = -1
  integer, parameter, public :: E3_DUPLICATE = -2
  public :: color_flow
  type :: color_flow
     type(color_state), dimension(:), allocatable :: cs
     type(epsilon3), dimension(:), allocatable :: e
  contains
     procedure :: check => color_flow_check     
     procedure :: to_string => color_flow_to_string
     procedure :: max_rank, max_rank_inflowing, max_rank_outflowing
     procedure :: contains_epsilons => color_flow_contains_epsilons
  end type color_flow
  public :: inflowing_lines, outflowing_lines, epsilon_lines
  integer, parameter, public :: CF_OK = 0
  integer, parameter, public :: CF_GHOST = -1
  integer, parameter, public :: CF_EPS = -2
  integer, parameter, public :: CF_SIZE_EPS = -3
  integer, parameter, public :: CF_DUPLICATE = -4
  integer, parameter, public :: CF_LINK = -5
  integer, parameter, public :: CF_ZERO = -6
contains
  impure elemental function color_state_check (cs) result (rc)
    class(color_state), intent(in) :: cs
    integer :: rc
    character(*), parameter :: my_name = "omega_birdtracks%color_state_check"
    type(color_state) :: cs_tmp    
    if (cs%g .and. (size (cs%i) > 0 .or. size (cs%o) > 0)) then
      cs_tmp = cs
      cs_tmp%g = .false.
      print *, my_name, ": ghost with open lines ", cs_tmp%to_string ()
      rc = CS_GHOST
    else
      rc = CS_OK
    end if
  end function color_state_check
  elemental function integer_width (i) result (w)
    integer, intent(in) :: i
    integer :: w
    character(IO_BUFFER_SIZE) :: buffer
    write (unit = buffer, fmt = "(I0)") i
    w = len_trim (buffer)
  end function integer_width
  pure function max_integer_width (a) result (w)
    integer, dimension(:), intent(in) :: a
    integer :: w
    w = max (integer_width (minval (a)), integer_width (maxval (a)))
  end function max_integer_width
  pure function integer_array_format (a, min_width, prefix, postfix) result (fmt)
    integer, dimension(:), intent(in) :: a
    integer, intent(in), optional :: min_width
    character(*), intent(in), optional :: prefix, postfix
    character(:), allocatable :: fmt
    character(:), allocatable :: pre, post
    character(IO_BUFFER_SIZE) :: buffer
    integer :: w
    w = max_integer_width (a)
    if (present (min_width)) then
       w = max (w, min_width)
    end if
    if (present (prefix)) then
      pre = prefix
    else
      pre = "["
    endif
    if (present (postfix)) then
      post = postfix
    else
      post = "]"
    endif
    select case (size (a))
    case (0)
       fmt = "'" // pre // post // "'"
    case (1)
       write (buffer, "(3A,I0,3A)") "'", pre, "', I", w, ", '", post, "'"
       fmt = trim (buffer)
    case default
       write (buffer, "(3A,3(I0,A),2A)") "'", pre, "',", size(a) - 1, "(I", w, ", ','), I", w, ", '", post, "'"
       fmt = trim (buffer)
    end select
  end function integer_array_format
  pure function color_state_to_string (cs) result (s)
    class(color_state), intent(in) :: cs
    character(:), allocatable :: s
    character(IO_BUFFER_SIZE) :: buffer
    integer :: iostat
    if (cs%g) then
      s = "g"
    else
      write (unit = buffer, fmt = &
               "('(', " // integer_array_format (cs%i) // ", ', ', " &
                        // integer_array_format (cs%o) // ", ')')",  &
             iostat = iostat) cs%i, cs%o
      if (is_iostat_eor (iostat) .or. is_iostat_end (iostat)) then
        s = "color_state_to_string: too long"
      else
        s = trim (buffer)
      endif
    end if
  end function color_state_to_string
  pure function color_state_conjugate (cs) result (csc)
    class(color_state), intent(in) :: cs
    type(color_state) :: csc
    csc = color_state (i = cs%o, o = cs%i, g = cs%g)
  end function color_state_conjugate
  elemental function rank_inflowing (cs) result (n)
    class(color_state), intent(in) :: cs
    integer :: n
    n = size (cs%i)
  end function rank_inflowing
  elemental function rank_outflowing (cs) result (n)
    class(color_state), intent(in) :: cs
    integer :: n
    n = size (cs%o)
  end function rank_outflowing
  elemental function color_state_contains_epsilons (cs) result (yorn)
    class(color_state), intent(in) :: cs
    logical :: yorn
    yorn = min (minval (cs%i), minval (cs%o)) < 0
  end function color_state_contains_epsilons
  impure elemental function epsilon3_check (e) result (rc)
    class(epsilon3), intent(in) :: e
    integer :: rc, n
    character(*), parameter :: my_name = "omega_birdtracks%epsilon3_check"    
    rc = E3_OK
    if (any (e%j >= 0)) then
      print *, my_name, ": index not negative in ", e%to_string ()
      rc = E3_NON_NEGATIVE
      return
    end if
    do n = 1, size (e%j)
      if (count (e%j == e%j(n)) /= 1) then
        print *, my_name, ": duplicate index in ", e%to_string ()
        rc = E3_DUPLICATE
        return
      end if
    end do
  end function epsilon3_check
  pure function epsilon3_to_string (e) result (s)
    class(epsilon3), intent(in) :: e
    character(:), allocatable :: s
    character(IO_BUFFER_SIZE) :: buffer
    integer :: iostat
    write (unit = buffer, fmt = &
               "('eps'," // integer_array_format (e%j, prefix = "(", postfix = ")") // ")",  &
           iostat = iostat) e%j
    if (is_iostat_eor (iostat) .or. is_iostat_end (iostat)) then
      s = "epsilon3_to_string: too long"
    else
      s = trim (buffer)
    endif
  end function epsilon3_to_string
  pure function inflowing_lines (cf, n_incoming) result (inflowing)
    class(color_flow), intent(in) :: cf
    integer, intent(in) :: n_incoming
    integer, dimension(:), allocatable :: inflowing
    integer :: i, j, n, num_lines
    num_lines = 0
    do n = 1, min (n_incoming, size (cf%cs))
      num_lines = num_lines + size (cf%cs(n)%o)
    end do
    do n = n_incoming + 1, size (cf%cs)
      num_lines = num_lines + size (cf%cs(n)%i)
    end do
    allocate (inflowing(num_lines))
    j = 0
    do n = 1, min (n_incoming, size (cf%cs))
      i = j + 1
      j = i + size (cf%cs(n)%o) - 1
      inflowing(i:j) = cf%cs(n)%o
    end do
    do n = n_incoming + 1, size (cf%cs)
      i = j + 1
      j = i + size (cf%cs(n)%i) - 1
      inflowing(i:j) = cf%cs(n)%i
    end do
  end function inflowing_lines
  pure function outflowing_lines (cf, n_incoming) result (outflowing)
    class(color_flow), intent(in) :: cf
    integer, intent(in) :: n_incoming
    integer, dimension(:), allocatable :: outflowing
    integer :: i, j, n, num_lines
    num_lines = 0
    do n = 1, min (n_incoming, size (cf%cs))
      num_lines = num_lines + size (cf%cs(n)%i)
    end do
    do n = n_incoming + 1, size (cf%cs)
      num_lines = num_lines + size (cf%cs(n)%o)
    end do
    allocate (outflowing(num_lines))
    j = 0
    do n = 1, min (n_incoming, size (cf%cs))
      i = j + 1
      j = i + size (cf%cs(n)%i) - 1
      outflowing(i:j) = cf%cs(n)%i
    end do
    do n = n_incoming + 1, size (cf%cs)
      i = j + 1
      j = i + size (cf%cs(n)%o) - 1
      outflowing(i:j) = cf%cs(n)%o
    end do
  end function outflowing_lines
  pure function epsilon_lines (cf) result (eps)
    class(color_flow), intent(in) :: cf
    integer, dimension(:), allocatable :: eps
    integer :: i, j, n, num_lines
    num_lines = 0
    do n = 1, size (cf%e)
      num_lines = num_lines + size (cf%e(n)%j)
    end do
    allocate (eps(num_lines))
    j = 0
    do n = 1, size (cf%e)
      i = j + 1
      j = i + size (cf%e(n)%j) - 1
      eps(i:j) = cf%e(n)%j
    end do
  end function epsilon_lines
  impure elemental function color_flow_check (cf) result (rc)
    class(color_flow), intent(in) :: cf
    integer :: rc, n
    integer, dimension(:), allocatable :: i, o, e
    integer :: n_incoming = 0    
    character(*), parameter :: my_name = "omega_birdtracks%color_flow_check"
    if (size (cf%cs) > 0) then
      rc = minval (cf%cs%check ())
      if (rc /= 0) then
        print *, my_name, ": inconsistent color_state in ", cf%to_string ()
      end if
    end if
    if (size (cf%e) > 0) then    
      rc = minval (cf%e%check ())
      if (rc /= 0) then
        print *, my_name, ": inconsistent epsilon in ", cf%to_string ()
        return
      end if
    end if
    i = inflowing_lines (cf, n_incoming)
    o = outflowing_lines (cf, n_incoming)
    e = epsilon_lines (cf)
    do n = 1, size (i)
      if (count (i == i(n)) /= 1) then
        print *, my_name, ": duplicate inflowing line ", i(n), " in ", cf%to_string ()
        rc = CF_DUPLICATE
        return
      end if
      if (i(n) > 0) then
        if (count (o == i(n)) /= 1) then
          print *, my_name, ": inflowing line ", i(n), " not uniquely connected in ", cf%to_string ()
          rc = CF_LINK
          return
        end if
      else if (i(n) < 0) then
        if (count (e == i(n)) /= 1) then
          print *, my_name, ": epsilon line ", i(n), " not uniquely connected in ", cf%to_string ()
          rc = CF_LINK
          return
        end if
      else
        print *, my_name, ": line zero in ", cf%to_string ()
        rc = CF_ZERO
        return
      end if
    end do
    do n = 1, size (o)
      if (count (o == o(n)) /= 1) then
        print *, my_name, ": duplicate outflowing line ", o(n), " in ", cf%to_string ()
        rc = CF_DUPLICATE
        return
      end if
      if (o(n) > 0) then
        if (count (i == o(n)) /= 1) then
          print *, my_name, ": outflowing line ", o(n), " not uniquely connected in ", cf%to_string ()
          rc = CF_LINK
          return
        end if
      else if (o(n) < 0) then
        if (count (e == o(n)) /= 1) then
          print *, my_name, ": epsilon line ", o(n), " not uniquely connected in ", cf%to_string ()
          rc = CF_LINK
          return
        end if
      else
        print *, my_name, ": line zero in ", cf%to_string ()
        rc = CF_ZERO
        return
      end if
    end do
    do n = 1, size (e)
      if (count (i == e(n)) + count (o == e(n)) /= 1) then
        print *, my_name, ": epsilon line ", e(n), " not uniquely connected in ", cf%to_string ()
        rc = CF_LINK
        return
      end if
    end do
  end function color_flow_check
  pure function color_flow_to_string (cf) result (s)
    class(color_flow), intent(in) :: cf
    character(:), allocatable :: s
    character(IO_BUFFER_SIZE) :: buffer
    integer :: n, from, to
    buffer = ""    
    if (size (cf%cs) > 0) then
      call initial (buffer, from, to, cf%cs(1)%to_string ())
      do n = 2, size (cf%cs)
        call append (buffer, from, to, ", ")
        call append (buffer, from, to, cf%cs(n)%to_string ())
      end do
      if (size (cf%e) > 0) then
        call append (buffer, from, to, "; ")
        call append (buffer, from, to, cf%e(1)%to_string ())
        do n = 2, size (cf%e)
          call append (buffer, from, to, ", ")
          call append (buffer, from, to, cf%e(n)%to_string ())
        end do
      end if
    end if
    s = "{" // trim (buffer) // "}"
  contains
    pure subroutine append (buffer, from, to, s)
      character(*), intent(inout) :: buffer
      integer, intent(inout) :: from, to
      character(*), intent(in) :: s
      from = to + 1
      to = from + len (s) - 1
      buffer(from:to) = s    
    end subroutine append
    pure subroutine initial (buffer, from, to, s)
      character(*), intent(inout) :: buffer
      integer, intent(inout) :: from
      integer, intent(out) :: to
      character(*), intent(in) :: s
      to = 0
      call append (buffer, from, to, s)
    end subroutine initial
  end function color_flow_to_string
  pure function max_rank_inflowing (cf) result (n)
    class(color_flow), intent(in) :: cf
    integer :: n
    n = maxval (cf%cs%rank_inflowing ())
  end function max_rank_inflowing
  pure function max_rank_outflowing (cf) result (n)
    class(color_flow), intent(in) :: cf
    integer :: n
    n = maxval (cf%cs%rank_outflowing ())
  end function max_rank_outflowing
  pure function max_rank (cf) result (n)
    class(color_flow), intent(in) :: cf
    integer :: n
    n = max (cf%max_rank_inflowing (), cf%max_rank_outflowing ())
  end function max_rank
  pure function color_flow_contains_epsilons (cf) result (yorn)
    class(color_flow), intent(in) :: cf
    logical :: yorn
    yorn = any (cf%cs%contains_epsilons ())
  end function color_flow_contains_epsilons
end module omega_birdtracks
